import Foundation
import Accelerate

struct Activation {
    let activate: (_: inout FloatBuffer) -> Void
    let gradient: (_: FloatBuffer) -> FloatBuffer
    
    static let sigmoid: Activation = Activation(activate: sigmoidActivation, gradient: sigmoidGradient)
}

func sigmoidActivation(buffer: inout FloatBuffer) {
    for i in 0..<buffer.count {
        buffer[i] = 1.0 / (1.0 + exp(-buffer[i]))
    }
}

func sigmoidGradient(buffer: FloatBuffer) -> FloatBuffer {
    var output = buffer
    var x: Float
    
    x = -1.0
    vDSP_vsmul(output, 1, &x, &output, 1, UInt(buffer.count))
    x = 1.0
    vDSP_vsadd(output, 1, &x, &output, 1, UInt(buffer.count))
    vDSP_vmul(buffer, 1, output, 1, &output, 1, UInt(buffer.count))
    return output
}
