import Foundation
import Accelerate

typealias Env = [String : [Float]]

typealias FloatBuffer = [Float]

class Node {
    let inputSize: Int
    let outputSize: Int
    
    init(inputSize: Int, outputSize: Int) {
        self.inputSize = inputSize
        self.outputSize = outputSize
    }
    
    func roll(env: Env) -> Instance {
        fatalError("abstract method must be implemented")
    }
}

class FilterNode: Node {
    let source: Node
    
    init(source: Node, outputSize: Int) {
        self.source = source
        
        super.init(inputSize: source.outputSize, outputSize: outputSize)
    }
    
    func activationGradient(for output: FloatBuffer) -> FloatBuffer {
        fatalError("abstract method must be implemented")
    }
}

class Instance {
    let previous: Instance?
    let node: Node
    
    var output: FloatBuffer
    
    init(of node: Node, from previous: Instance?, output: FloatBuffer? = nil) {
        self.previous = previous
        self.node = node
        if let output = output {
            self.output = output
        } else {
            self.output = FloatBuffer(repeating: 0.0, count: node.outputSize)
        }
    }
}
