import Foundation
import Accelerate

class LossNode: FilterNode {
    let loss: Loss
    let name: String
    
    init(source: Node, loss: Loss = Loss.l2, name: String) {
        self.loss = loss
        self.name = name
        
        super.init(source: source, outputSize: source.outputSize)
    }
    
    override func roll(env: Env) -> Instance {
        guard let expected = env[name] else {
            fatalError("missing expected output in environment")
        }
        
        guard expected.count == source.outputSize else {
            fatalError("output size mismatch")
        }
        
        let previous = self.source.roll(env: env)

        let lossInput = (expected: expected, actual: previous.output)
        let lossValue = loss.loss(lossInput)
        let output = loss.gradient(lossInput)

        let instance = LossInstance(of: self, from: previous, loss: lossValue, output: output)
        return instance
    }
}

class LossInstance: Instance {
    let loss: Float
    
    init(of node: Node, from previous: Instance?, loss: Float, output: FloatBuffer?) {
        self.loss = loss
        
        super.init(of: node, from: previous, output: output)
    }
}
