import Foundation

public enum FileError: Error {
    case invalidType
    case invalidLength
}

public func loadLabelFile(filename: String) throws -> [Int] {
    let url = URL.init(fileURLWithPath: "./\(filename)")
    let blob = try Data(contentsOf: url)
    let size = try blob.withUnsafeBytes{ (pointer: UnsafePointer<UInt32>) -> Int in
        guard pointer.pointee.byteSwapped == 0x801 else {
            throw FileError.invalidType
        }
        
        let sizePointer = pointer.successor()
        return Int(sizePointer.pointee.byteSwapped)
    }
    
    guard blob.count == 8 + size else {
        throw FileError.invalidLength
    }
    
    var labels = [Int]()
    labels.reserveCapacity(size)
    for byte in blob.subdata(in: 8..<blob.count) {
        labels.append(Int(byte))
    }
    return labels
}

public typealias RawImage = [UInt8]

public struct Images {
    public let width: Int
    public let height: Int
    public let rawImages: [RawImage]
}

public func loadImageFile(filename: String) throws -> Images {
    let url = URL.init(fileURLWithPath: "./\(filename)")
    let blob = try Data(contentsOf: url)
    
    let (size, width, height) = try blob.withUnsafeBytes{ (pointer: UnsafePointer<UInt32>) -> (size: Int, width: Int, height: Int) in
        guard pointer.pointee.byteSwapped == 0x803 else {
            throw FileError.invalidType
        }
        
        let sizePointer = pointer.successor()
        let size = Int(sizePointer.pointee.byteSwapped)
        let heightPointer = sizePointer.successor()
        let height = Int(heightPointer.pointee.byteSwapped)
        let widthPointer = heightPointer.successor()
        let width = Int(widthPointer.pointee.byteSwapped)
        
        return (size: size, width: width, height: height)
    }
    
    var imageStart = 16
    let imageSize = width * height
    var imagesLeft = size
    var rawImages = [RawImage]()
    
    repeat {
        let rawImage = RawImage(blob.subdata(in: imageStart..<imageStart + imageSize))
        rawImages.append(rawImage)
        
        imageStart += imageSize
        imagesLeft -= 1
    } while imagesLeft > 0
    
    return Images(width: width, height: height, rawImages: rawImages)
}
