import Foundation

func runBatch(on graph: Node,
              input: [FloatBuffer], inputLabel: String = "input",
              output: [FloatBuffer], outputLabel: String = "expected",
              samples: [Int]) -> Float {
    guard samples.count > 0 else {
        return 0.0
    }
    
    let array = UnsafeMutablePointer<Instance>.allocate(capacity: samples.count)
    DispatchQueue.concurrentPerform(iterations: samples.count) { i in
        let env = [inputLabel: input[samples[i]], outputLabel: output[samples[i]]]
        let pointer = array.advanced(by: i)
        pointer.initialize(to: graph.roll(env: env))
    }
    var instances = [Instance?]((0..<samples.count).lazy.map { array.advanced(by: $0).pointee })
    array.deinitialize(count: samples.count)
    array.deallocate()

    var lossSum: Float = 0.0
    
    while instances.first! != nil {
        if let instance = (instances.first!)! as? FullyConnectedGradientFilterInstance {
            var slopeGrads = [FloatBuffer]()
            slopeGrads.reserveCapacity(instances.count)
            var biasGrads = [FloatBuffer]()
            biasGrads.reserveCapacity(instances.count)
            
            for instance in instances {
                let instance = instance! as! FullyConnectedGradientFilterInstance
                slopeGrads.append(instance.gradient)
                biasGrads.append(instance.error)
            }
            
            (instance.node as! FullyConnectedGradientFilterNode).connected.weight.update(slopeGrads: slopeGrads, biasGrads: biasGrads)
        } else if let _ = (instances.first!)! as? LossInstance {
            for instance in instances {
                let instance = instance! as! LossInstance
                lossSum += instance.loss
            }
        }
        
        for i in 0..<instances.count {
            instances[i] = instances[i]!.previous
        }
    }
    
    return lossSum / Float(samples.count)
}
