import Foundation

let inputLayerWeight = InMemoryWeight(inputSize: 28 * 28, outputSize: 64, learningRate: 0.05)
let hiddenLayerWeight = InMemoryWeight(inputSize: 64, outputSize: 32, learningRate: 0.05)
let outputLayerWeight = InMemoryWeight(inputSize: 32, outputSize: 10, learningRate: 0.05)

let image = ImageNode(outputSize: 28 * 28, name: "input")

let inputLayer = FullyConnectedFilterNode(source: image, weight: inputLayerWeight)
let hiddenLayer = FullyConnectedFilterNode(source: inputLayer, weight: hiddenLayerWeight)
let outputLayer = FullyConnectedFilterNode(source: hiddenLayer, weight: outputLayerWeight)

let loss = LossNode(source: outputLayer, loss: Loss.crossEntropy, name: "expected")

let outputGradient = outputLayer.gradientFilter(source: loss)
let hiddenGradient = hiddenLayer.gradientFilter(source: outputGradient)
let inputGradient = inputLayer.gradientFilter(source: hiddenGradient)

let predictGraph = outputLayer
let trainGraph = inputGradient

let labels = try loadLabelFile(filename: "train-labels.idx1-ubyte")
let images = try loadImageFile(filename: "train-images.idx3-ubyte")
let testLabels = try loadLabelFile(filename: "t10k-labels.idx1-ubyte")
let testImages = try loadImageFile(filename: "t10k-images.idx3-ubyte")

let inputs: [[Float]] = images.rawImages.map{ $0.map{ Float($0) / 256.0 } }
let outputs: [[Float]] = labels.map{ x in
    var array = [Float](repeating: 0.0, count: 10)
    array[x] = 1.0
    return array
}

let testInputs: [[Float]] = testImages.rawImages.map{ $0.map{ Float($0) / 256.0 } }
let testOutputs: [[Float]] = testLabels.map{ x in
    var array = [Float](repeating: 0.0, count: 10)
    array[x] = 1.0
    return array
}

func argmax(_ xs: [Float]) -> Int {
    var i = 0
    var max = xs[i]
    for j in 1..<xs.count {
        if xs[j] > max {
            i = j
            max = xs[j]
        }
    }
    return i
}

let batchPerEpoch = 1000
let totalEpoch = 100

var samples = [Int](repeating: 0, count: 64)
for i in 0..<batchPerEpoch * totalEpoch {
    for j in 0..<samples.count {
        samples[j] = Int.random(in: 0..<inputs.count)
    }

    let loss = runBatch(on: trainGraph, input: inputs, output: outputs, samples: samples)

    if i % batchPerEpoch == 0 {
        var correct = 0
        for _ in 0..<512 {
            let j = Int.random(in: 0..<testInputs.count)

            let env = ["input": testInputs[j], "expected": testOutputs[j]]
            let instance = predictGraph.roll(env: env)
            
            if argmax(instance.output) == testLabels[j] {
                correct += 1
            }
        }
        let rate = Float(correct) / 512.0

        print("batch \(i): rate = \(rate * 100.0), loss = \(loss)")
    }
}
